let op_table_sz = 17
let clause_table_sz = 17
let ctx_sz = 17
let const_table_sz = 17
let domain_table_sz = 17
let zinc_expr_sz = 17

let string_of_float' f =
  let s = string_of_float f in
  if String.sub s (String.length s - 1) 1 = "." then
    s ^ "0"
  else
    s
    
module StringIntHashtbl = Hashtbl.Make (struct
  type t = string * int

  let equal = (=)

  let hash = Hashtbl.hash
end)
    
let suffix_from s i = String.sub s i (String.length s - i)
  
let rec split_from c s i =
  match
    try
      Some (String.index_from s i c)
    with Not_found ->
      None
  with
  | None -> [suffix_from s i]
  | Some j ->
      String.sub s i (j - i) :: split_from c s (succ j)

let split c s =
  split_from c s 0

let clp2zincpath =
  try
    split ':' (Sys.getenv "CLP2ZINCPATH")
  with Not_found ->
    []
   
let search_path = ref clp2zincpath

let prefix_path = ref ["."]
    
let () =
  search_path := !search_path @ [Filename.concat (Filename.dirname Sys.argv.(0)) "lib"]
    
let list_of_queue queue =
  let dequeue list item =
    item :: list in
  let rev_list = Queue.fold dequeue [] queue in
  List.rev rev_list
     
type term = term' * Lexer.position
and term' =
  | Var of string
  | Int of int
  | Float of float
  | String of string
  | Record of (string * term) list
  | Term of string * term array

let term_true pos = (Term ("true", [| |]), pos)
        
type entry =
  | Minizinc of string
  | Include of string * Lexer.position
  | Clause of string * term array * term
  | Goal of term
  | Var_decl of term

exception End_of_file

module StringHashtbl = Hashtbl.Make (struct
  type t = string
  
  let hash = Hashtbl.hash

  let equal = ( = )
end)

type bin_op =
  | Non_assoc
  | Left_assoc
  | Right_assoc
  | Both_assoc

type unary_op =
  | Unary_non_assoc
  | Unary_assoc
      
type right_op =
  | Bin_op of bin_op
  | Unary_op of unary_op

type op =
  | Left of string * int * unary_op * Lexer.position
  | Bin of string * int * bin_op * Lexer.position
    
type op_table = {
    left_ops: (int * unary_op) StringHashtbl.t;
    right_ops: (int * right_op) StringHashtbl.t;
  }

let op op_table p a symbol =
  match a with
  | "xfx" ->
      StringHashtbl.add op_table.right_ops symbol (p, Bin_op Non_assoc)
  | "yfx" ->
      StringHashtbl.add op_table.right_ops symbol (p, Bin_op Left_assoc)
  | "xfy" ->
      StringHashtbl.add op_table.right_ops symbol (p, Bin_op Right_assoc)
  | "yfy" ->
      StringHashtbl.add op_table.right_ops symbol (p, Bin_op Both_assoc)
  | "xf" ->
      StringHashtbl.add op_table.right_ops symbol (p, Unary_op Unary_non_assoc)
  | "yf" ->
      StringHashtbl.add op_table.right_ops symbol (p, Unary_op Unary_assoc)
  | "fx" ->
      StringHashtbl.add op_table.left_ops symbol (p, Unary_non_assoc)
  | "fy" ->
      StringHashtbl.add op_table.left_ops symbol (p, Unary_assoc)
  | _ ->
      failwith "Unknown op spec"
      
let create_op_table () =
  let left_ops = StringHashtbl.create op_table_sz in
  let right_ops = StringHashtbl.create op_table_sz in
  let op_table = { left_ops; right_ops } in
  op op_table 400 "yfx" "div";
  op op_table 400 "yfx" "mod";
  op op_table 700 "xfx" ">=";
  op op_table 400 "yfx" "*";
  op op_table 200 "fy" "+";
  op op_table 500 "yfy" "+";
  op op_table 500 "yfy" "++";
  op op_table 1000 "xfy" ",";
  op op_table 200 "fy" "-";
  op op_table 200 "fy" "not";
  op op_table 500 "yfx" "-";
  op op_table 400 "yfx" "/";
  op op_table 1100 "xfy" ";";
  op op_table 700 "xfx" "=";
  op op_table 700 "xfx" "==";
  op op_table 700 "xfx" "\\=";
  op op_table 700 "xfx" "\\==";
  op op_table 700 "xfx" "!=";
  op op_table 700 "xfx" "<=";
  op op_table 700 "xfx" ">";
  op op_table 700 "xfx" ">=";
  op op_table 700 "xfx" "<";
  op op_table 900 "fy" "\\+";
  op op_table 200 "xfy" "^";
  op op_table 700 "xfx" "=..";
  op op_table 925 "yfy" "/\\";
  op op_table 950 "yfy" "\\/";
  op op_table 1200 "fx" ":-";
  op op_table 1200 "xfx" ":-";
  op op_table 1050 "yfx" "->";
  op op_table 1150 "yfx" "<->";
  op op_table 550 "xfy" "..";
  op op_table 570 "xfx" "in";
  op op_table 600 "fy" "var";
  op op_table 650 "xfy" "of";
  op op_table 700 "xfx" ":";
  op_table
    
let check_dot (final, position) =
   if final <> Lexer.Atom "." then
    Lexer.error position "Clause should end with a dot"

let check_par_open (final, position) =
  if final <> Lexer.Par_open then
    Lexer.error position "`(' expected"

let check_par_close (final, position) =
  if final <> Lexer.Par_close then
    Lexer.error position "`)' expected"

let check_bracket_close (final, position) =
  if final <> Lexer.Bracket_close then
    Lexer.error position "`]' expected"

let check_curly_close (final, position) =
  if final <> Lexer.Curly_close then
    Lexer.error position "`}' expected"

let rec make_list_tail pos list term =
  match list with
  | [] ->
      term
  | head :: tail ->
      (Term (".", [| head; make_list_tail pos tail term |]), pos)

let rec extract_record_fields term =
  match fst term with
  | Term (",", [| a; b |]) ->
      begin
        match extract_record_fields a with
        | None -> None
        | Some a ->
            match extract_record_fields b with
            | None -> None
            | Some b -> Some (a @ b)
      end
  | Term (":", [| Term (field, [| |]), _; v |]) ->
      Some [(field, v)]
  | _ ->
      None


let rec make_record term =
  match extract_record_fields term with
  | None ->
      term
  | Some fields ->
      (Record fields, snd term)

let apply_op op value_stack =
  match op with
  | Left (f, _, _, pos) ->
      begin
        match value_stack with
        | arg :: values ->
            (Term (f, [| arg |]), pos) :: values
        | _ ->
            assert false
      end
  | Bin (f, _, _, pos) ->
      begin
        match value_stack with
        | right :: left :: values ->
            (Term (f, [| left; right |]), pos) :: values
        | _ ->
            assert false
      end
        
let rec flush_ops value_stack op_stack =
  match op_stack with
  | [] ->
      begin
        match value_stack with
        | [v] -> v
        | _ -> failwith ("flush_ops " ^ string_of_int (List.length value_stack))
      end
  | op :: ops ->
      flush_ops (apply_op op value_stack) ops

let left_assoc op =
  match op with
  | Bin_op Left_assoc
  | Bin_op Both_assoc
  | Unary_op Unary_assoc ->
      true
  | Bin_op Non_assoc
  | Bin_op Right_assoc
  | Unary_op Unary_non_assoc ->
      false
        
let has_priority op' p op f =
  match op' with
  | Left (f', p', desc, pos) ->
      if p' < p then
        true
      else if p < p' then
        false
      else if desc = Unary_assoc && not (left_assoc op) then
        true
      else if desc = Unary_non_assoc && left_assoc op then
        false
      else if f = f' then
        Lexer.error pos (Printf.sprintf "non-composable unary operator `%s'" f)
      else
        Lexer.error pos (Printf.sprintf "incompatible operators `%s' and `%s'" f f')
  | Bin (f', p', desc, pos) ->
      if p' < p then
        true
      else if p < p' then
        false
      else if (desc = Left_assoc || desc = Both_assoc || desc = Non_assoc) && left_assoc op then
        true
      else if desc = Right_assoc && not (left_assoc op) then
        false
      else if f = f' then
        Lexer.error pos (Printf.sprintf "non-associative operator `%s'" f)
      else
        Lexer.error pos (Printf.sprintf "incompatible operators `%s' and `%s'" f f')
        
let rec apply_ops value_stack op_stack p op f =
  match op_stack with
  | op' :: ops when has_priority op' p op f ->
      apply_ops (apply_op op' value_stack) ops p op f
  | _ ->
      value_stack, op_stack
    
let rec parse_term op_table lexbuf =
  parse_left op_table [] ([] : op list) false lexbuf
and parse_left op_table value_stack op_stack in_list lexbuf =
  parse_left_term op_table value_stack op_stack in_list lexbuf (Lexer.term lexbuf)
and parse_left_term op_table value_stack op_stack in_list lexbuf term =
  match term with
  | Lexer.Atom_dot f ->
      (Term (f, [| |]), Lexer.position lexbuf), (Lexer.Atom ".", Lexer.position lexbuf)
  | Lexer.Atom_par f ->
      let args, final = parse_term_list op_table lexbuf in
      check_par_close final;
      parse_right op_table ((Term (f, Array.of_list args), Lexer.position lexbuf) :: value_stack) op_stack in_list lexbuf
  | Lexer.Atom (("forall" | "exists") as f) ->
      check_par_open (Lexer.term lexbuf, Lexer.position lexbuf);
      let q, final = parse_term op_table lexbuf in
      check_par_close final;
      check_par_open (Lexer.term lexbuf, Lexer.position lexbuf);
      let term, final = parse_term op_table lexbuf in
      check_par_close final;parse_right op_table ((Term (f, [| q; term |]), Lexer.position lexbuf) :: value_stack) op_stack in_list lexbuf
  | Lexer.Atom f ->
      begin
        try
          let p, op = StringHashtbl.find op_table.left_ops f in
          parse_left op_table value_stack (Left (f, p, op, Lexer.position lexbuf) :: op_stack) in_list lexbuf
        with
        | Not_found ->
            parse_right op_table ((Term (f, [| |]), Lexer.position lexbuf) :: value_stack) op_stack in_list lexbuf
      end
  | Lexer.Int i ->
      parse_right op_table ((Int i, Lexer.position lexbuf) :: value_stack) op_stack in_list lexbuf
  | Lexer.Float f ->
      parse_right op_table ((Float f, Lexer.position lexbuf) :: value_stack) op_stack in_list lexbuf
  | Lexer.String s ->
      parse_right op_table ((String s, Lexer.position lexbuf) :: value_stack) op_stack in_list lexbuf
  | Lexer.Var x ->
      parse_right op_table ((Var x, Lexer.position lexbuf) :: value_stack) op_stack in_list lexbuf
  | Lexer.Par_open ->
      let term, final = parse_term op_table lexbuf in
      check_par_close final;
      parse_right op_table (make_record term :: value_stack) op_stack in_list lexbuf
  | Lexer.Curly_open ->
      let term, final = parse_term op_table lexbuf in
      check_curly_close final;
      parse_right op_table (make_record term :: value_stack) op_stack in_list lexbuf
  | Lexer.Bracket_open ->
      let list, (final, position) = parse_term_list op_table lexbuf in
      let tail =
        match final with
        | Lexer.Bracket_close -> 
            Term ("[]", [| |]), position
        | Lexer.Atom "|" ->
            let term, final = parse_term op_table lexbuf in
            check_bracket_close final;
            term
        | _ -> Lexer.error position "`]' or `|' expected" in
      parse_right op_table (make_list_tail position list tail :: value_stack) op_stack in_list lexbuf
  | Lexer.Double_bracket_open ->
      let rec parse_lines accu =
        let list, (final, position) = parse_term_list op_table lexbuf in
        match final with
        | Lexer.Double_bracket_close ->
            let lines = List.rev (list :: accu) in
            let empty_list = Term ("[]", [| |]), position in
            let lines' = List.map (fun line -> make_list_tail position line empty_list) lines in
            make_list_tail position lines' empty_list
        | Lexer.Atom "|" ->
            parse_lines (list :: accu)
        | _ -> Lexer.error position "`|]' or `|' expected" in
      let m = parse_lines [] in
      parse_right op_table (m :: value_stack) op_stack in_list lexbuf
  | lexem ->
      Lexer.error (Lexer.position lexbuf) "Left symbol expected"
and parse_right op_table value_stack op_stack in_list lexbuf =
  let term = Lexer.term lexbuf in
  match term with
  | Lexer.Par_close | Lexer.Bracket_close | Lexer.Double_bracket_close | Lexer.Curly_close | Lexer.Atom "." ->
      flush_ops value_stack op_stack, (term, Lexer.position lexbuf)
  | Lexer.Atom f ->
      begin
        if in_list && (f = "," || f = "|" || f = ";") then
          flush_ops value_stack op_stack, (term, Lexer.position lexbuf)
        else
          try
            let p, op = StringHashtbl.find op_table.right_ops f in
            let value_stack, op_stack = apply_ops value_stack op_stack p op f in
            match op with
            | Bin_op op ->
                parse_left op_table value_stack (Bin (f, p, op, Lexer.position lexbuf) :: op_stack) in_list lexbuf
            | Unary_op _ ->
                begin
                  match value_stack with
                  | v :: value_stack ->
                      parse_right op_table ((Term (f, [| v |]), Lexer.position lexbuf) :: value_stack) op_stack in_list lexbuf
                  | _ ->
                      failwith "value_stack"
                end
          with
          | Not_found ->
              Lexer.error (Lexer.position lexbuf) "Operator expected"
      end
  | Lexer.Projector f ->
      begin
        let value_stack, op_stack = apply_ops value_stack op_stack 400 (Unary_op Unary_assoc) "xfx" in
        match value_stack with
        | v :: value_stack ->
            parse_right op_table ((Term (":", [| v; Term (f, [| |]), Lexer.position lexbuf |]), Lexer.position lexbuf) :: value_stack) op_stack in_list lexbuf
        | _ ->
            failwith "value_stack"
      end
  | Lexer.Bracket_open ->
      let list, final = parse_term_list op_table lexbuf in
      check_bracket_close final;
      begin
        match value_stack with
        | v :: value_stack ->
            parse_right op_table ((Term ("[]", Array.of_list (v :: list)), Lexer.position lexbuf) :: value_stack) op_stack in_list lexbuf
        | _ ->
            failwith "value_stack"
      end
  | l ->
      Lexer.error (Lexer.position lexbuf) "right symbol expected"
and parse_term_list op_table lexbuf =
  match Lexer.term lexbuf with
  | Lexer.Bracket_close ->
      [], (Lexer.Bracket_close, Lexer.position lexbuf)
  | term ->
      let head, final = parse_left_term op_table [] [] true lexbuf term in
      parse_term_tail op_table lexbuf [head] final
and parse_term_tail op_table lexbuf accu (term, position) =
  match term with
  | Lexer.Atom "," ->
      let term, final = parse_left op_table [] [] true lexbuf in
      parse_term_tail op_table lexbuf (term :: accu) final
  | Lexer.Atom "|" | Lexer.Bracket_close | Lexer.Double_bracket_close | Lexer.Par_close ->
      List.rev accu, (term, position)
  | _ ->
      Lexer.error position "`,' or `|' expected"

let const_table = StringHashtbl.create const_table_sz
    
let check_semicolon position final =
  if final = Lexer.Atom ";" then
    ()
  else
    Lexer.error position "`;' expected"
       
module Int = struct
  type t = int

  let compare = compare
end
 
module IntMap = Map.Make (Int)

module Value = struct
  type t = t' * Lexer.position
  and t' =
    | Var of int * bool ref list ref
    | Int of int
    | Float of float
    | String of string
    | Record of (string * t) list
    | Term of string * t array

  let rec subst binder s t =
    match t with
    | Var _, _
    | Int _, _
    | Float _, _
    | String _, _ ->
        t
    | Record l, pos ->
        Record (List.map (fun (label, v) -> (label, subst binder s v)) l), pos
    | Term (f, [| |]), pos when f = binder ->
        s
    | Term (f, args), pos ->
        Term (f, Array.map (subst binder s) args), pos

  exception List_expected

  let rec to_list_exc g =
    match g with
    | Term ("[]", [| |]), _ ->
        []
    | Term (".", [| hd; tl |]), _ ->
        hd :: to_list_exc tl
    | _ ->
        raise List_expected

  let rec to_list g =
    try
      to_list_exc g
    with List_expected ->
      match g with
      | _, pos ->
          Lexer.error pos "list expected"
          
  let var_counter = ref 0

  let fresh pos =
    incr var_counter;
    Var (!var_counter, ref []), pos

  let rec deref env value =
    match fst value with
    | Var (i, _) ->
        begin
          try
            deref env (IntMap.find i env)
          with Not_found ->
            value
        end
     | _ -> value

  let rec ground env x =
    match fst (deref env x) with
    | Var _ -> false
    | Int _
    | Float _
    | String _ -> true
    | Record fields -> List.for_all (fun (f, v) -> ground env v) fields
    | Term (_, args) -> List.for_all (ground env) (Array.to_list args)

  let rec deep_deref env value =
    match deref env value with
    | Term (f, args), pos ->
        Term (f, Array.map (deep_deref env) args), pos
    | v -> v

  exception Conversion

  let of_atom pos s =
    Term (s, [| |]), pos

  let of_bool pos b =
    if b then of_atom pos "true"
    else of_atom pos "false"

  let to_atom_conv env value =
    match fst (deref env value) with
    | Term (s, [| |]) -> s
    | _ -> raise Conversion

  let to_atom env value =
    try
      to_atom_conv env value
    with Conversion ->
      Lexer.error (snd value) "atom expected"

  let to_int_conv env value =
    match fst (deref env value) with
    | Int i -> i
    | _ -> raise Conversion

  let to_int env value =
    try
      to_int_conv env value
    with Conversion ->
      Lexer.error (snd value) "integer value expected"
          
  let to_float_conv env value =
    match fst (deref env value) with
    | Int i -> float_of_int i
    | Float f -> f
    | _ -> raise Conversion

  let to_float env value =
    try
      to_float_conv env value
    with Conversion ->
      Lexer.error (snd value) "number expected"
 
  let op_table = create_op_table ()

  let parentheses p o out_chan x =
    if p then
      Format.fprintf out_chan "(%a)" o x
    else
      Format.fprintf out_chan "%a" o x
      
  let rec output env out_chan value =
    output_prec env 1300 out_chan value
  and output_prec env prec out_chan value =
    let rec output_tail out_chan value =
      match deref env value with
      | Term (".", [| a; b |]), _ ->
          Format.fprintf out_chan ", %a%a" (output_prec env 999) a output_tail b
      | Term ("[]", [| |]), _ ->
          ()
      | _ ->
          Format.fprintf out_chan " | %a" (output env) value in
    let output_term f args =
      try
        if Array.length args = 1 then
          match StringHashtbl.find op_table.left_ops f with
          | (prec', assoc) ->
              parentheses (prec < prec')
                (fun out_chan arg ->
                  Format.fprintf out_chan "%s %a" f
                    (output_prec env (if assoc = Unary_assoc then prec' else pred prec')) arg)
                out_chan args.(0)
        else if Array.length args = 2 then
          match StringHashtbl.find op_table.right_ops f with
          | (prec', Bin_op assoc) ->
              let f =
                if f = ".." then ".."
                else if f = ":" then ": "
                else if f = "," then ", "
                else " " ^ f ^ " " in
              parentheses (prec < prec')
                (fun out_chan () ->
                  Format.fprintf out_chan "%a%s%a"
                    (output_prec env (if assoc = Left_assoc || assoc = Both_assoc then prec' else pred prec')) args.(0) f
                    (output_prec env (if assoc = Right_assoc || assoc = Both_assoc then prec' else pred prec')) args.(1)) out_chan ()
          | _ ->
              raise Not_found
        else
          raise Not_found
      with Not_found ->
        let output_args out_chan args =
          for i = 0 to Array.length args - 1 do
            if (i >= 1) then
              Format.fprintf out_chan ", ";
            Format.fprintf out_chan "%a" (output_prec env 999) args.(i);
          done in
        Format.fprintf out_chan "%s(%a)" f output_args args in
    let output_matrix m =
      let rec output_list_tail out_chan l =
        match l with
        | [] ->
            ()
        | hd :: tl ->
            Format.fprintf out_chan ", %a%a" (output env) hd output_list_tail tl in
      let output_list out_chan l =
        match l with
        | [] ->
            Format.fprintf out_chan ""
        | hd :: tl ->
            Format.fprintf out_chan "%a%a" (output env) hd output_list_tail tl in
      let rec output_matrix out_chan value =
        match value with
        | hd :: tl ->      
            Format.fprintf out_chan "| %a %a" output_list hd output_matrix tl
        | _ ->
            () in
      Format.fprintf out_chan "[%a|]" output_matrix m in
    match deref env value with
    | Var (x, _), _ ->
        Format.fprintf out_chan "_%d" x
    | Term (a, [| |]), _ ->
        Format.fprintf out_chan "%s" a
    | Int i, _ ->
        Format.fprintf out_chan "%d" i
    | Float f, _ ->
        Format.fprintf out_chan "%s" (string_of_float' f)
    | String s, _ ->
        Format.fprintf out_chan "\"%s\"" s
    | Record fields, _ ->
        let output_field out_chan (f, v) =
          Format.fprintf out_chan "%s: %a" f (output_prec env 500) v in
        let rec output_fields out_chan fields =
          match fields with
          | [] -> ()
          | [field] -> output_field out_chan field
          | head :: tail -> Format.fprintf out_chan "%a, %a" output_field head output_fields tail in
        Format.fprintf out_chan "(%a)" output_fields fields
    | (Term (".", [| a; b |]), _) as m ->
        begin
          try
            output_matrix (List.map to_list_exc (to_list_exc m))
          with List_expected ->
            Format.fprintf out_chan "[%a%a]" (output env) a output_tail b
        end
    | Term ("[]", args), _ ->
        if Array.length args = 0 then
          Format.fprintf out_chan "[]"
        else
          let output_args out_chan args =
            if Array.length args >= 1 then
              begin
                output env out_chan args.(1);
                for i = 2 to Array.length args - 1 do
                  Format.fprintf out_chan ", %a" (output_prec env 999) args.(i);
                done
              end in
          Format.fprintf out_chan "%a[%a]" (output env) args.(0) output_args args
    | Term ("matrix", [| a |]), _ ->
        begin
          try
            output_matrix (List.map to_list_exc (to_list_exc a))
          with List_expected ->
            output_term "matrix" [| a |]
        end
    | Term ((("forall" | "exists") as f), [| q; term |]), _ ->
        Format.fprintf out_chan "%s (%a) (%a)" f (output env) q (output env) term
    | Term ("-", [| a |]), _ ->
        parentheses (prec > 100) (output_prec env 100) out_chan a
    | Term (f, args), _ ->
        output_term f args

  let to_string env v =
    let buff = Buffer.create 16 in
    Format.fprintf (Format.formatter_of_buffer buff) "%a@?" (output env) v;
    Buffer.contents buff
          
  let unify_var i v env =
    IntMap.add i v env

  let rec mark m env v =
    match fst (deref env v) with
    | Var (_, l) ->
        l := m :: !l
    | Int _
    | Float _
    | String _ -> ()
    | Record fields ->
        List.iter (fun (f, v) -> mark m env v) fields
    | Term (_, args) ->
        Array.iter (mark m env) args

  let wake_up l =
    List.iter (fun m -> m := true) !l

  let of_list pos l =
    List.fold_right (fun hd tl -> Term (".", [| hd; tl |]), pos) l (Term ("[]", [| |]), pos)
end
        
let rec verbatim_value_of_term term =
  match term with
  | Var v, pos -> Value.Term (v, [| |]), pos
  | Int i, pos -> Value.Int i, pos
  | Float f, pos -> Value.Float f, pos
  | String s, pos -> Value.String s, pos
  | Record r, pos -> Value.Record (List.map (fun (f, v) -> (f, verbatim_value_of_term v)) r), pos
  | Term (f, args), pos ->
      Value.Term (f, Array.map verbatim_value_of_term args), pos

let format_term out_chan term =
  Value.output IntMap.empty out_chan (verbatim_value_of_term term)

let parse_entry op_table lexbuf =
  match Lexer.main lexbuf with
  | Lexer.Atom ":-" ->
      let goal, final = parse_term op_table lexbuf in
      check_dot final;
      Goal goal
  | Lexer.Atom "var"
  | Lexer.Atom "array" as term ->
      let decl, (final, position) = parse_left_term op_table [] [] true lexbuf term in
      let () = check_semicolon position final in
      Var_decl decl
  | Lexer.Atom_dot f ->
      Clause (f, [| |], term_true (Lexer.position lexbuf))
  | Lexer.Atom_turnstile f ->
      let goal, final = parse_term op_table lexbuf in
      check_dot final;
      Clause (f, [| |], goal)
  | Lexer.Atom_equal f ->
      let decl, (final, position) = parse_left op_table [] ([] : op list) true lexbuf in
      let () = check_semicolon position final in
      StringHashtbl.add const_table f decl;
      let buff = Buffer.create 16 in
      Format.fprintf (Format.formatter_of_buffer buff) "%s = %a;@?" f format_term decl;
      Minizinc (Buffer.contents buff)
  | Lexer.Atom_par f ->
      let args, final = parse_term_list op_table lexbuf in
      check_par_close final;
      let args = Array.of_list args in
      begin
        match Lexer.term lexbuf with
        | Lexer.Atom ":-" ->
            let goal, final = parse_term op_table lexbuf in
            check_dot final;
            Clause (f, args, goal)
        | Lexer.Atom "." ->
            Clause (f, args, (term_true (Lexer.position lexbuf)))
        | _ ->
            Lexer.error (Lexer.position lexbuf) "Clause body expected"
      end
  | Lexer.Minizinc contents ->
      Minizinc contents
  | Lexer.Include filename ->
      if Filename.check_suffix filename ".plz" then
        Include (filename, Lexer.position lexbuf)
      else
        Minizinc (Printf.sprintf "include \"%s\";" filename)
  | Lexer.End_of_file ->
      raise End_of_file
  | _ ->
      assert false

let open_lexing filename =
  let in_chan = open_in filename in
  let lexbuf = Lexing.from_channel in_chan in
  lexbuf.Lexing.lex_curr_p <- { lexbuf.Lexing.lex_curr_p with Lexing.pos_fname = filename };
  in_chan, lexbuf
        
let rec parse' op_table accu lexbuf =
  match
    try
      Some (parse_entry op_table lexbuf)
    with
      End_of_file -> None
  with
  | None ->
      List.rev accu
  | Some (Include (filename, pos)) ->
      let in_chan, lexbuf' =
        try open_lexing filename
        with Sys_error msg ->
          let rec search list =
            match list with
            | [] ->
                Lexer.error pos msg
            | path :: tail ->
                begin
                  let rec search_prefix list =
                    match list with
                    | [] ->
                        search tail
                    | prefix :: tail ->
                        begin
                          let prefixed_filename =
                            if prefix = "." then filename
                            else Filename.concat prefix filename in
                          try open_lexing (Filename.concat path prefixed_filename)
                          with Sys_error _ ->
                            search_prefix tail
                        end in
                  search_prefix !prefix_path
                end in
          search !search_path in
      let entries = parse lexbuf' in
      close_in in_chan;
      parse' op_table (List.rev_append entries accu) lexbuf
  | Some entry ->
      parse' op_table (entry :: accu) lexbuf
and parse lexbuf =
  let op_table = create_op_table () in
  parse' op_table [] lexbuf


let load_file filename =
  let in_chan, lexbuf =
    try open_lexing filename
    with Sys_error msg ->
      prerr_endline msg;
      exit 1 in
  let entries = parse lexbuf in
  close_in in_chan;
  entries
    
let asserta clauses f args body =
  let key = f, Array.length args in
  let argsbody = args, body in
  try
    StringIntHashtbl.replace clauses key (argsbody :: StringIntHashtbl.find clauses key)
  with Not_found->
    StringIntHashtbl.add clauses key [argsbody]

let expect_colon term =
  match term with
  | Term (":", [| ty; var |]), _ ->
      ty, var
  | _, pos ->
      Lexer.error pos "declaration of the shape `type: var' expected"

let expect_ident term =
  match term with
  | Term (ident, [| |]), _ ->
      ident
  | _, pos ->
      Lexer.error pos "identifier expected"
        
let fresh_var_counter = ref 0

let fresh_var () =
  incr fresh_var_counter;
  Printf.sprintf "X%d" !fresh_var_counter

let int_of_bool b =
  if b then
    1
  else
    0
    
let output_neg_cond out_chan (x, v) =
  Printf.fprintf out_chan "%s = %d" x (int_of_bool (not v))
    
let rec output_neg_conds_tail out_chan conds =
  match conds with
  | [] ->
      ()
  | head :: tail ->
      Printf.fprintf out_chan " \\/ %a%a" output_neg_cond head output_neg_conds_tail tail
  
let output_neg_conds out_chan conds =
  match conds with
  | [] ->
      ()
  | head :: tail ->
      Printf.fprintf out_chan "%a%a -> " output_neg_cond head output_neg_conds_tail tail
    
let output_cond out_chan (x, v) =
  Printf.fprintf out_chan "%s = %d" x (int_of_bool v)
    
let rec output_conds_tail out_chan conds =
  match conds with
  | [] ->
      ()
  | head :: tail ->
      Printf.fprintf out_chan " /\\ %a%a" output_cond head output_conds_tail tail
  
let output_conds out_chan conds =
  match conds with
  | [] ->
      ()
  | head :: tail ->
      Printf.fprintf out_chan "%a%a -> " output_cond head output_conds_tail tail
        
module IntSet = Set.Make (Int)
        
module Ground = struct
  type t = t' * Lexer.position
  and t' =
    | Int of int
    | Float of float
    | String of string
    | Record of (string * t) list
    | Term of string * t array

  exception List_expected

  let rec to_list_exc g =
    match g with
    | Term ("[]", [| |]), _ ->
        []
    | Term (".", [| hd; tl |]), _ ->
        hd :: to_list_exc tl
    | _ ->
        raise List_expected

  let rec to_list g =
    try
      to_list_exc g
    with List_expected ->
      match g with
      | _, pos ->
          Lexer.error pos "list expected"

  let of_list pos l =
    List.fold_right (fun hd tl -> Term (".", [| hd; tl |]), pos) l (Term ("[]", [| |]), pos)
          
  let rec compare a b =
    match fst a, fst b with
    | Term (f, args), Term (f', args') ->
        let o = String.compare f f' in
        if o = 0 then
          let o = Pervasives.compare (Array.length args) (Array.length args') in
          if o = 0 then
            let rec compare_args i =
              if i < Array.length args then
                let o = compare args.(i) args'.(i) in
                if o = 0 then
                  compare_args (succ i)
                else
                  o
              else
                0 in
            compare_args 0
          else
            o
        else
          o
    | a, b ->
        Pervasives.compare a b
          
  let rec of_value env value =
    match Value.deref env value with
    | Value.Term (f, args), pos ->
        Term (f, Array.map (of_value env) args), pos
    | Value.Record fields, pos ->
        Record (List.map (fun (f, v) -> (f, of_value env v)) fields), pos
    | Value.Int i, pos -> Int i, pos
    | Value.Float f, pos -> Float f, pos
    | Value.String s, pos -> String s, pos
    | Value.Var _, pos ->
        Lexer.error pos "free variable unallowed"

  let rec to_value v =
    match v with
    | Term (f, args), pos -> Value.Term (f, Array.map to_value args), pos
    | Int i, pos -> Value.Int i, pos
    | Float f, pos -> Value.Float f, pos
    | String s, pos -> Value.String s, pos
    | Record r, pos -> Value.Record (List.map (fun (f, v) -> (f, to_value v)) r), pos
          
  let op_table = create_op_table ()

  let output out_chan value =
    Value.output IntMap.empty out_chan (to_value value)
end
        
let rec value_of_term ctx term =
  match term with
  | Var v, pos ->
      if v = "_" then
        Value.fresh pos
      else
        begin
          try
            StringHashtbl.find ctx v
          with Not_found ->
            let value = Value.fresh pos in
            StringHashtbl.add ctx v value;
            value
        end
  | Int i, pos -> Value.Int i, pos
  | Float f, pos -> Value.Float f, pos
  | String s, pos -> Value.String s, pos
  | Record r, pos -> Value.Record (List.map (fun (f, v) -> (f, value_of_term ctx v)) r), pos
  | Term (f, args), pos ->
      Value.Term (f, Array.map (value_of_term ctx) args), pos

module Domain = struct
  type t =
    | Array of int * int * t
    | Matrix of int * int * int * int * t
    | Int of int * int
    | RangeFloat of float * float
    | Float
    | Bool

  let rec get_int pos dom =
    match dom with
    | Array (_, _, dom') -> get_int pos dom'
    | Matrix (_, _, _, _, dom') -> get_int pos dom'
    | Int (min, max) -> (min, max)
    | _ ->
        Lexer.error pos "integer domain expected"

  let rec output out_chan dom =
    match dom with
    | Array (lb, ub, dom') ->
        Format.fprintf out_chan "array[%d..%d] of %a" lb ub output dom'
    | Matrix (lb0, ub0, lb1, ub1, dom') ->
        Format.fprintf out_chan "array[%d..%d, %d..%d] of %a" lb0 ub0 lb1 ub1 output dom'
    | Int (min, max) ->
        Format.fprintf out_chan "var %d..%d" min max
    | RangeFloat (min, max) ->
        Format.fprintf out_chan "var %s..%s" (string_of_float' min) (string_of_float' max)
    | Float ->
        Format.fprintf out_chan "var float"
    | Bool ->
        Format.fprintf out_chan "var bool"
end

let rec extract_list list =
  match list with
  | Value.Term ("..", [| Value.Int i, _; Value.Int j, _ |]), pos ->
      let rec make_range v accu =
        if v >= i then
          make_range (pred v) ((Value.Int v, pos) :: accu)
        else
          accu in
      let range = make_range j [] in
      Some range
  | Value.Term ("[]", [| |]), _ ->
      Some []
  | Value.Term (".", [| hd; tl |]), _ ->
      begin
        match extract_list tl with
        | None -> None
        | Some tl -> Some (hd :: tl)
      end
  | _ ->
      None

let round f =
  floor (f +. 0.5)
      
let debug = ref false

let expand_rules = ref false  

let eval_arithmetic domain_table clauses env term =
  if !debug then
    Format.eprintf "evaluate %a@." (Value.output env) term;
  let rec eval env accu term =
    let eval_args args =
      let add_arg (accu, accu_arg) arg =
        let accu, arg = eval env accu arg in
        (accu, arg :: accu_arg) in
      let accu, accu_args = List.fold_left add_arg (accu, []) args in
      accu, List.rev accu_args in
    let bin_op pos i f s a b =
      let accu, a = eval env accu a in
      let accu, b = eval env accu b in
      let v =
        match fst a, fst b with
        | Value.Int a, Value.Int b ->
            Value.Int (i a b), pos
        | Value.Int a, Value.Float b ->
            Value.Float (f (float_of_int a) b), pos
        | Value.Float a, Value.Int b ->
            Value.Float (f a (float_of_int b)), pos
        | Value.Float a, Value.Float b ->
            Value.Float (f a b), pos
        | _ ->
            Value.Term (s, [| a; b |]), pos in
      accu, v in
    let bin_compare pos i f s a b =
      let accu, a = eval env accu a in
      let accu, b = eval env accu b in
      let v =
        match fst a, fst b with
        | Value.Int a, Value.Int b ->
            Value.of_bool pos (i a b)
        | Value.Int a, Value.Float b ->
            Value.of_bool pos (f (float_of_int a) b)
        | Value.Float a, Value.Int b ->
            Value.of_bool pos (f a (float_of_int b))
        | Value.Float a, Value.Float b ->
            Value.of_bool pos (f a b)
        | _ ->
            Value.Term (s, [| a; b |]), pos in
      accu, v in
    let float_op pos f s a b =
      let accu, a = eval env accu a in
      let accu, b = eval env accu b in
      let v =
        try
          Value.Float (f (Value.to_float_conv env a) (Value.to_float_conv env b)), pos
        with Value.Conversion ->
          Value.Term (s, [| a; b |]), pos in
      accu, v in
    let float_to_int f s a =
      let accu, a = eval env accu a in
      let v =
        match a with
        | Value.Int a, pos ->
            Value.Int a, pos
        | Value.Float a, pos ->
            Value.Int (int_of_float (f a)), pos
        | _, pos ->
            Value.Term (s, [| a |]), pos in
      accu, v in
    let float_fun f s a =
      let accu, a = eval env accu a in
      let v =
        match a with
        | Value.Int a, pos ->
            Value.Float (f (float_of_int a)), pos
        | Value.Float a, pos ->
            Value.Float (f a), pos
        | _, pos ->
            Value.Term (s, [| a |]), pos in
      accu, v in
    let indexical f i f_bounds x args pos =
      let d0 = fresh_var () in
      let x = Value.deref env x in
      match x with
      | Value.Var _, pos
      | Value.Term (".", [| _; _ |]), pos ->
          accu, (Value.Term (f, Array.of_list (x :: args)), pos)
      | _ ->
          let x = Ground.of_value env x in
          let name =
            match x with
            | Ground.Term (name, [| |]), _ ->
                name
            | Ground.Term ("[]", [| (Ground.Term (name, [| |]), _); _ |]), _
            | Ground.Term ("[]", [| (Ground.Term (name, [| |]), _); _; _ |]), _ ->
                name
            | _, pos ->
                Lexer.error pos "model variable expected" in
          let bounds =
            try
              Domain.get_int (snd x) (StringHashtbl.find domain_table name)
            with Not_found ->
              Lexer.error (snd x) (Format.sprintf "unknown domain variable `%s'" name) in
          let eval_args args =
            let add_arg (accu, accu_arg) arg =
              let accu, arg = eval env accu arg in
              let arg = Ground.of_value env arg in
              (accu, arg :: accu_arg) in
            let accu, accu_args = List.fold_left add_arg (accu, []) args in
            accu, List.rev accu_args in
          let accu, args = eval_args args in
          let (lb, ub) = f_bounds bounds in
          StringHashtbl.add domain_table d0 (Domain.Int (lb, ub));
          (Ground.Term (i, Array.of_list ((Ground.Term (d0, [| |]), pos) :: x :: args)), pos) :: accu,
          (Value.Term (d0, [| |]), pos) in
    match Value.deref env term with
    | Value.Record fields, pos ->
        let add_field (accu, rev_fields) (f, v) =
          let accu, v = eval env accu v in
          accu, (f, v) :: rev_fields in
        let accu, rev_fields =
          List.fold_left add_field (accu, []) fields in
        accu, (Value.Record (List.rev fields), pos)
    | Value.Term (":", [| r; v |]), pos ->
        let accu, r = eval env accu r in
        begin
          match r with
          | Value.Record fields, _ ->
              begin
                match v with
                | Value.Term (f, [| |]), _ ->
                    begin
                      try
                        accu, List.assoc f fields
                      with Not_found ->
                        Lexer.error (snd v) ("no field labeled with " ^ f)
                    end
                | _ ->
                    Lexer.error (snd v) ("field name expected")
              end
          | _ ->
              accu, term
        end
    | Value.Term ("+", [| a; b |]), pos ->
        bin_op pos ( + ) ( +. ) "+" a b
    | Value.Term ("-", [| a; b |]), pos ->
        bin_op pos ( - ) ( -. ) "-" a b
    | Value.Term ("-", [| a |]), pos ->
        let accu, a = eval env accu a in
        let v =
          match fst a with
          | Value.Int a ->
              Value.Int (- a), pos
          | Value.Float a ->
              Value.Float (-. a), pos
          | _ ->
              Value.Term ("-", [| a |]), pos in
        accu, v
    | Value.Term ("ceil", [| a |]), pos ->
        float_to_int ceil "ceil" a
    | Value.Term ("floor", [| a |]), pos ->
        float_to_int floor "floor" a
    | Value.Term ("round", [| a |]), pos ->
        float_to_int round "round" a
    | Value.Term ("sin", [| a |]), pos ->
        float_fun sin "sin" a
    | Value.Term ("cos", [| a |]), pos ->
        float_fun cos "cos" a
    | Value.Term ("atan", [| a |]), pos ->
        float_fun atan "atan" a
    | Value.Term ("*", [| a; b |]), pos ->
        bin_op pos ( * ) ( *. ) "*" a b
    | Value.Term ("div", [| a; b |]), pos ->
        bin_op pos ( / ) ( /. ) "div" a b
    | Value.Term ("rem", [| a; b |]), pos ->
        bin_op pos ( mod ) ( mod_float ) "rem" a b
    | Value.Term ("log", [| a; b |]), pos ->
        float_op pos (fun a b -> log b /. log a) "log" a b
    | Value.Term ("/", [| a; b |]), pos ->
        float_op pos ( /. ) "/" a b
    | Value.Term ("=", [| a; b |]), pos ->
        bin_compare pos ( = ) ( = ) "=" a b
    | Value.Term ("<=", [| a; b |]), pos ->
        bin_compare pos ( <= ) ( <= ) "<=" a b
    | Value.Term (">", [| a; b |]), pos ->
        bin_compare pos ( > ) ( > ) ">" a b
    | Value.Term (">=", [| a; b |]), pos ->
        bin_compare pos ( >= ) ( >= ) ">=" a b
    | Value.Term ("<", [| a; b |]), pos ->
        bin_compare pos ( < ) ( < ) "<" a b
    | Value.Term ("++", [| a; b |]), pos ->
        let accu, a = eval env accu a in
        let accu, b = eval env accu b in
        let v =
          match a, b with
          | (Value.String a, _), (Value.String b, _) ->
              Value.String (a ^ b), pos
          | _ ->
	      match extract_list a, extract_list b with
	      | Some la, Some lb ->
		  Value.of_list pos (la @ lb)
	      | _ ->
		  Value.Term ("++", [| a; b |]), pos in
        accu, v
    | Value.Term ("\\/", [| a; b |]), pos ->
        let accu, a = eval env accu a in
        let accu, b = eval env accu b in
        let v =
          match a, b with
          | (Value.Term ("true", [| |]), _ as v), _
          | _, (Value.Term ("true", [| |]), _ as v) ->
              v
          | (Value.Term ("false", [| |]), _), v
          | v, (Value.Term ("false", [| |]), _) ->
              v
          | _ ->
              Value.Term ("\\/", [| a; b |]), pos in
        accu, v
    | ((Value.Int _ | Value.Float _), _) as v ->
        accu, v
    | Value.Term ("min", [| a; b |]), pos ->
        bin_op pos ( min ) ( min ) "min" a b
    | Value.Term ("max", [| a; b |]), pos ->
        bin_op pos ( max ) ( max ) "max" a b
    | Value.Term ("pow", [| a; b |]), pos ->
        float_op pos ( ** ) "pow" a b
    | Value.Term ("min", [| x |]), pos ->
        indexical "min" "indexical_min" (fun bounds -> bounds) x [] pos
    | Value.Term ("max", [| x |]), pos ->
        indexical "max" "indexical_max" (fun bounds -> bounds) x [] pos
    | Value.Term ("card", [| x |]), pos ->
        indexical "card" "indexical_card" (fun (lb, ub) -> (1, ub - lb + 1)) x [] pos
    | Value.Term ("dom_nth", [| x; n |]), pos ->
        indexical "dom_nth" "indexical_dom_nth" (fun bounds -> bounds) x [n] pos
    | Value.Term (".", [| e; r |]), pos ->
        let accu, e = eval env accu e in
        let accu, r = eval env accu r in
        let accu, v =
          match r with
          | Value.Term ("in", [| Value.Term (x, [| |]), _; Value.Term ("..", [| Value.Int i, _; Value.Int j, _ |]), _ |]), _ ->
	      let rec mk_list i j =
		if i <= j then
		  (Value.subst x (Value.Int i, pos) e) :: mk_list (succ i) j
		else
		  [] in
              accu, Value.of_list pos (mk_list i j)
          | _ ->
              accu, (Value.Term (".", [| e; r |]), pos) in
        accu, v
    | Value.Term ("[]", [| l; i |]), pos ->
        let accu, l = eval env accu l in
        let accu, i = eval env accu i in
        let accu, v =
          match i, extract_list l with
          | (Value.Int i, _), Some l when List.length l >= i ->
              accu, List.nth l (i - 1)
          | _ ->
              accu, (Value.Term ("[]", [| l; i |]), pos) in
        accu, v
    | Value.Term ("[]", [| m; i; j |]), pos ->
        let accu, m = eval env accu m in
        let accu, i = eval env accu i in
        let accu, j = eval env accu j in
        begin
          let extract_from_matrix l =
            match i, extract_list l with
            | (Value.Int i, _), Some l when List.length l >= i ->
                begin
                  match j, extract_list (List.nth l (i - 1)) with
                  | (Value.Int j, _), Some l when List.length l >= j ->
                      Some (accu, List.nth l (j - 1))
                  | _ ->
                      None
                end
            | _ ->
                None in
          match
            match m with
            | Value.Term ("matrix", [| l |]), _ ->
                extract_from_matrix l
            | _ ->
                extract_from_matrix m
          with
          | Some (accu, v) ->
              accu, v
          | None ->
              accu, (Value.Term ("[]", [| m; i; j |]), pos)
        end
    | Value.Term ("name", [| x |]), pos ->
        begin
          let s = Value.to_string env x in
          accu, (Value.String s, pos)
        end
    | Value.Term ((("forall" | "exists") as f), [| Value.Term ("in", [| Value.Term (binder, [| |]), _; list |]), _; term |]), pos ->
        begin
          let accu, list = eval env accu list in
          let accu, term = eval env accu term in
          let accu, v =
            match extract_list list with
            | None ->
                accu, (Value.Term (f, [| Value.Term ("in", [| Value.Term (binder, [| |]), pos; list |]), pos; term |]), pos)
            | Some list ->
                let operator, neutral =
                  if f = "forall" then "/\\", "true"
                  else "\\/", "false" in
                let compose a b =
                  Value.Term (operator, [| a; Value.subst binder b term |]), pos in
                let v = List.fold_left compose (Value.Term (neutral, [| |]), pos) list in
                eval env accu v in
          accu, v
        end
    | Value.Term (f, args), pos ->
        let accu, args = eval_args (Array.to_list args) in
        let arity = List.length args in
        begin
          match
            if !expand_rules then
              match clauses with
              | None -> None
              | Some clauses ->
                  try
                    let f_clauses = StringIntHashtbl.find clauses (f, arity) in
                    let distinct_variables args =
                      let previous = Hashtbl.create 17 in
                      let distinct v =
                        match v with
                        | Var v, _ ->
                            if Hashtbl.mem previous v then
                              false
                            else
                              begin
                                Hashtbl.add previous v ();
                                true
                              end
                        | _ -> false in
                      List.for_all distinct (Array.to_list args) in
                    match f_clauses with
                    | [(args, body)] when distinct_variables args -> Some (args, body)
                    | _ -> None
                  with Not_found -> None
            else
              None
          with
          | Some (args', body) ->
              let subst_table = Hashtbl.create 17 in
              let args = Array.of_list args in
              for i = 0 to Array.length args - 1 do
                match args'.(i) with
                | Var v, _ ->
                    Hashtbl.add subst_table v args.(i)
                | _ ->
                    raise Exit
              done;
              let rec subst term =
                match term with
                | Var v, pos ->
                    begin
                      try
                        Hashtbl.find subst_table v
                      with Not_found ->
                        Lexer.error pos ("unknown variable '" ^ v ^ "'")
                    end
                | Term (f, args), pos ->
                    Value.Term (f, Array.map subst args), pos
                | Record fields, pos ->
                    Value.Record (List.map (fun (f, v) -> (f, subst v)) fields), pos
                | Int i, pos -> Value.Int i, pos
                | Float f, pos -> Value.Float f, pos
                | String s, pos -> Value.String s, pos in
              eval env accu (subst body)
          | None ->
              begin
                match
                  if arity = 0 then
                    try
                      Some (StringHashtbl.find const_table f)
                    with Not_found ->
                      None
                  else
                    None
                with
                | Some v -> eval env accu (value_of_term (StringHashtbl.create 17) v)
                | None -> accu, (Value.Term (f, Array.of_list args), pos)
              end
        end
    | k ->
        accu, k in
  let accu, v = eval env [] term in
  (List.rev accu, v)
        
type decision_tree = decision_tree' * Lexer.position
and decision_tree' =
  | False
  | True
  | And of decision_tree list
  | Or of decision_tree list
  | Search of Ground.t
  | Constraint of Ground.t * bool

let rec has_equiv t =
  match fst t with
  | False
  | True
  | Search _ -> false
  | And list
  | Or list -> List.exists has_equiv list
  | Constraint (_, equiv) -> equiv
        
let rec output_decision_tree out_chan t =
  let print_sub_nodes out_chan l =
    List.iter (output_decision_tree out_chan) l in
  match fst t with
  | False -> Format.fprintf out_chan "[.\\node{$\\bot$};]"
  | True -> Format.fprintf out_chan "[.\\node{$\\top$};]"
  | And l ->
      Format.fprintf out_chan "[.\\node{$\\wedge$};%a]" print_sub_nodes l
  | Or l ->
      Format.fprintf out_chan "[.\\node{$\\vee$};%a]" print_sub_nodes l
(*  | Search (Ground.Term ("indexical_min", [| m; x |]), _) ->
      Format.fprintf out_chan "[.\\node{$\\text{\\lstinline|%a|} \\leftarrow \\min(\\text{\\lstinline|%a|})$};]" Ground.output m Ground.output x
  | Search (Ground.Term ("indexical_max", [| m; x |]), _) ->
      Format.fprintf out_chan "[.\\node{$\\text{\\lstinline|%a|} \\leftarrow \\max(\\text{\\lstinline|%a|})$};]" Ground.output m Ground.output x*)
  | Search t
  | Constraint (t, _) ->
      Format.fprintf out_chan "[.\\node{\\lstinline|%a|};]" Ground.output t
    
let mk_and pos list =
  let rec filter accu list =
    match list with
    | [] -> Some accu
    | (False, _) :: _ -> None
    | (True, _) :: tail -> filter accu tail
    | (And sub_list, _) :: tail ->
        begin
          match filter accu sub_list with
          | None -> None
          | Some accu -> filter accu tail
        end
    | item :: tail -> filter (item :: accu) tail in
  match filter [] list with
  | None -> False, pos
  | Some [] -> True, pos
  | Some [item] -> item
  | Some list -> And (List.rev list), pos

let mk_or pos list =
  let rec filter accu list =
    match list with
    | [] -> accu
    | (False, _) :: tail -> filter accu tail
    | (Or sub_list, _) :: tail when not (List.exists has_equiv sub_list) -> filter (filter accu sub_list) tail
    | item :: tail -> filter (item :: accu) tail in
  match filter [] list with
  | [] -> False, pos
  | [item] when not (has_equiv item) -> item
  | list ->
      let list = List.rev list in
      Or list, pos

let goal_of_clauses f args clauses pos =
  match clauses with
  | [] -> Value.Term ("false", [| |]), pos
  | head :: tail ->
      let mk_term (args', body) =
        let clause = value_of_term (StringHashtbl.create ctx_sz) (Term (",", [| Term (f, args'), pos; body |]), pos) in
        let (args', body) =
          match clause with
          | Value.Term (",", [| Value.Term (_, args'), _; body |]), _ ->
              (args', body)
          | _ ->
              failwith "mk_term" in
        Value.Term (",", [| Value.Term ("=", [| Value.Term (f, args), pos; Value.Term (f, args'), pos |]), pos; body |]), pos in
      List.fold_left (fun x y -> Value.Term (";", [| x; mk_term y |]), pos) (mk_term head) tail

let global_vars = StringHashtbl.create 17

let global_bvars = StringHashtbl.create 17
        
let dummy_linenumber = { Lexer.line = 0; Lexer.column = 0}
    
let dummy_pos = { Lexer.filename = ""; start_p = dummy_linenumber; end_p = dummy_linenumber }
    
exception Unify

let zinc_expr = StringIntHashtbl.create zinc_expr_sz

let () =
  StringIntHashtbl.add zinc_expr ("+", 2) ();
  StringIntHashtbl.add zinc_expr ("+", 1) ();
  StringIntHashtbl.add zinc_expr ("-", 2) ();
  StringIntHashtbl.add zinc_expr ("-", 1) ();
  StringIntHashtbl.add zinc_expr ("*", 2) ();
  StringIntHashtbl.add zinc_expr ("/", 2) ();
  StringIntHashtbl.add zinc_expr ("div", 2) ();
  StringIntHashtbl.add zinc_expr ("rem", 2) ();
  StringIntHashtbl.add zinc_expr ("<->", 2) ();
  StringIntHashtbl.add zinc_expr ("->", 2) ();
  StringIntHashtbl.add zinc_expr ("\\/", 2) ();
  StringIntHashtbl.add zinc_expr ("/\\", 2) ();
  StringIntHashtbl.add zinc_expr ("not", 1) ();
  StringIntHashtbl.add zinc_expr ("floor", 1) ();
  StringIntHashtbl.add zinc_expr ("ceil", 1) ();
  StringIntHashtbl.add zinc_expr ("round", 1) ();
  StringIntHashtbl.add zinc_expr ("min", 2) ();
  StringIntHashtbl.add zinc_expr ("max", 2) ();
  StringIntHashtbl.add zinc_expr ("int2float", 1) ();
  StringIntHashtbl.add zinc_expr ("matrix", 1) ();
  StringIntHashtbl.add zinc_expr ("[]", 2) ();
  StringIntHashtbl.add zinc_expr ("[]", 3) ()

let is_zinc_expr domain_table env e f arity =
  if not (Value.ground env e) then
    false
  else if arity = 0 then
    StringHashtbl.mem domain_table f
  else
    StringIntHashtbl.mem zinc_expr (f, arity)

let rec range i j =
  if i <= j then
    i :: range (succ i) j
  else
    []

let make_matrix x pos min0 max0 min1 max1 =
  let get_nth n0 n1 =
    Value.Term ("[]", [| Value.Term (x, [| |]), pos; Value.Int n0, pos; Value.Int n1, pos |]), pos in
  let get_line n =
    Value.of_list pos (List.map (get_nth n) (range min1 max1)) in
  Value.of_list pos (List.map get_line (range min0 max0))
    
let unify test domain_table v1 v2 env =
  let rec unify v1 v2 (env, accu) =
    let v1 = Value.deref env v1 in
    let v2 = Value.deref env v2 in
    if !debug then
      Format.eprintf "try to unify %a and %a@." (Value.output env) v1 (Value.output env) v2;
    let put_equal_constraint () =
      if test then raise Unify;
      (env, (Constraint ((Ground.Term ("=", [| Ground.of_value env v1; Ground.of_value env v2 |]), snd v1), false), snd v1) :: accu) in
    match v1, v2 with
    | (Value.Var (i1, l1), _), (Value.Var (i2, l2), _) ->
        if i1 > i2 then
          begin
            if test then raise Unify;
            Value.wake_up l1;
            (Value.unify_var i1 v2 env, accu)
          end
        else if i2 > i1 then
          begin
            if test then raise Unify;
            Value.wake_up l2;
            (Value.unify_var i2 v1 env, accu)
          end
        else
          (env, accu)
    | (Value.Var (i, l), _), v
    | v, (Value.Var (i, l), _) ->
        if test then raise Unify;
        Value.wake_up l;
        (Value.unify_var i v env, accu)
    | (Value.Int i1, _), (Value.Int i2, _) ->
        if i1 = i2 then
          (env, accu)
        else
          raise Unify
    | (Value.Float f1, _), (Value.Float f2, _) ->
        if f1 = f2 then
          (env, accu)
        else
          raise Unify
    | (Value.String s1, _), (Value.String s2, _) ->
        if s1 = s2 then
          (env, accu)
        else
          raise Unify
    | (Value.Term (".", [| head; tail |]), _ as l), (Value.Term (x, [| |]), pos)
    | (Value.Term (x, [| |]), pos), (Value.Term (".", [| head; tail |]), _ as l) ->
        begin
          match
            try
              StringHashtbl.find domain_table x
            with Not_found ->
              raise Unify
          with
          | Domain.Array (min, max, dom) ->
              let get_nth n =
                Value.Term ("[]", [| Value.Term (x, [| |]), pos; Value.Int n, pos |]), pos in
              let l' = Value.of_list pos (List.map get_nth (range min max)) in
              unify l l' (env, accu)
          | Domain.Matrix (min0, max0, min1, max1, dom) ->
              unify l (make_matrix x pos min0 max0 min1 max1) (env, accu)
          | _ ->
              raise Unify
        end
    | (Value.Term (".", [| head; tail |]), _ as l), (Value.Term ("matrix", [| l' |]), pos)
    | (Value.Term ("matrix", [| l' |]), pos), (Value.Term (".", [| head; tail |]), _ as l) ->
        unify l l' (env, accu)
    | (Value.Term ("matrix", [| l |]), _), (Value.Term (x, [| |]), pos)
    | (Value.Term (x, [| |]), pos), (Value.Term ("matrix", [| l |]), _) ->
        begin
          match
            try
              StringHashtbl.find domain_table x
            with Not_found ->
              raise Unify
          with
          | Domain.Matrix (min0, max0, min1, max1, dom) ->
              unify l (make_matrix x pos min0 max0 min1 max1) (env, accu)
          | _ ->
              raise Unify
        end
    | ((Value.Int _ | Value.Float _), _), ((Value.Term (f, args), _) as v)
    | ((Value.Term (f, args), _) as v), ((Value.Int _ | Value.Float _), _)->
        if is_zinc_expr domain_table env v f (Array.length args) then
          put_equal_constraint ()
        else
          raise Unify
    | (Value.Term (f1, args1), _), (Value.Term (f2, args2), _) ->
        let arity1 = Array.length args1
        and arity2 = Array.length args2 in
        if f1 = f2 && arity1 = 0 && arity2 = 0 then
          (env, accu)
        else if is_zinc_expr domain_table env v1 f1 arity1 && is_zinc_expr domain_table env v2 f2 arity2 then
          put_equal_constraint ()
        else if f1 = f2 && arity1 = arity2 then
          let ctx = ref (env, accu) in
          for i = 0 to arity1 - 1 do
            ctx := unify args1.(i) args2.(i) !ctx
          done;
          !ctx
        else
          raise Unify
    | _ ->
        raise Unify in
  let (env, accu) = unify v1 v2 (env, []) in
  (env, List.rev accu)

let rec same_goal domain_table env a b =
  match Value.deref env a, Value.deref env b with
  | (Value.Var _, _), (Value.Var _, _) ->
      true
  | (Value.Int i, _), (Value.Int j, _) ->
      i = j
  | (Value.Float f, _), (Value.Float g, _) ->
      f = g
  | (Value.String s, _), (Value.String t, _) ->
      s = t
  | (Value.Term (f1, args1), _ as v1), (Value.Term (f2, args2), _ as v2) ->
      let arity1 = Array.length args1 in
      let arity2 = Array.length args2 in
      if is_zinc_expr domain_table env v1 f1 arity1 && is_zinc_expr domain_table env v2 f2 arity2 then
        true
      else if f1 = f2 && arity1 = arity2 then
        try
          for i = 0 to arity1 - 1 do
            if not (same_goal domain_table env args1.(i) args2.(i)) then
              raise Exit
          done;
          true
        with Exit ->
          false
      else
        false
  | _ ->
      false

let goal_true env = [| env |], (fun nodes -> nodes.(0))

let goal_seq pos env s = [| env |], (fun nodes -> mk_and pos (s @ [nodes.(0)]))

let goal_false pos = [| |], (fun nodes -> False, pos)

let make_search pos list =
  List.map (fun s -> Search s, pos) list
    
let bool pos env s b =
  if b then
    goal_seq pos env s
  else
    goal_false pos

let close_true pos (anchors, close) =
  close (Array.map (fun _ -> True, pos) anchors)
        
let evaluate_goal domain_table clauses goal =
  let rec eval history env k_exp goal =
    let eval_unify pos a b env orders =
      let ordersa, a = eval_arithmetic domain_table None env a in
      let ordersb, b = eval_arithmetic domain_table None env b in
      try
        let env, cstrs = unify false domain_table a b env in
        goal_seq pos env (orders @ make_search pos (ordersa @ ordersb) @ cstrs)
      with Unify ->
        goal_false pos in
    let test_equal pos a b =
      let ordersa, a = eval_arithmetic domain_table None env a in
      let ordersb, b = eval_arithmetic domain_table None env b in
      let equal =
        try
          ignore (unify true domain_table a b env);
          true
        with Unify ->
          false in
      make_search pos (ordersa @ ordersb), equal in
    let check_equal pos a b =
      let s, equal = test_equal pos a b in
      bool pos env s equal in
    let check_not_equal pos a b =
      let s, equal = test_equal pos a b in
      bool pos env s (not equal) in
    let check_order pos s o_i o_f a b =
      let ordersa, a = eval_arithmetic domain_table None env a in
      let ordersb, b = eval_arithmetic domain_table None env b in
      let o, constraints =
        try
          o_i (Value.to_int_conv env a) (Value.to_int_conv env b), []
        with Value.Conversion ->
          try
            o_f (Value.to_float_conv env a) (Value.to_float_conv env b), []
          with Value.Conversion ->
            true, [Constraint ((Ground.Term (s, [| Ground.of_value env a; Ground.of_value env b |]), pos), false), pos] in
      let s = make_search pos (ordersa @ ordersb) @ constraints in
      bool pos env s o in
    let eval_args clauses pos args =
      let orders_args = Array.map (eval_arithmetic domain_table clauses env) args in
      let orders = make_search pos (List.flatten (List.map fst (Array.to_list orders_args))) in
      let args = Array.map snd orders_args in
      orders, args in
    let goal = Value.deref env goal in
    if !debug then
      Format.eprintf "execute %a@." (Value.output env) goal;
    match goal with
    | Value.Term ("true", [| |]), _ ->
        goal_true env
    | Value.Term ("false", [| |]), pos ->
        goal_false pos
    | Value.Term ("\\+", [| a |]), pos ->
        begin
          match eval history env k_exp a with
          | [| |], _ -> goal_true env
          | _, _ -> goal_false pos
        end
    | Value.Term (",", [| a; b |]), pos ->
        let k_exp_a = (Value.Term (",", [| b; k_exp |]), pos) in
        let anchors_a, close_a = eval history env k_exp_a a in
        let anchors_list_ref = ref [] in
        let close_list_ref = ref [] in
        let offset_ref = ref 0 in
        for i = 0 to Array.length anchors_a - 1 do
          let anchors_b, close_b = eval history anchors_a.(i) k_exp b in
          let anchors_b_length = Array.length anchors_b in
          anchors_list_ref := anchors_b :: !anchors_list_ref;
          let offset = !offset_ref in
          close_list_ref := (close_b, offset, anchors_b_length) :: !close_list_ref;
          offset_ref := offset + anchors_b_length;
        done;
        let anchors_list = List.rev !anchors_list_ref in
        let close_list = List.rev !close_list_ref in
        let anchors = Array.concat anchors_list in
        let close nodes =
          let extract_sub (close_b, offset, length) =
            close_b (Array.sub nodes offset length) in
          let nodes_a_list = List.map extract_sub close_list in
          let nodes_a = Array.of_list nodes_a_list in
          close_a nodes_a in
        anchors, close
    | Value.Term (";", [| a; b |]), pos ->
        let m = ref false in
        Value.mark m env k_exp;
        let anchors_a, close_a = eval history env k_exp a in
        let anchors_b, close_b = eval history env k_exp b in
        if anchors_a = [| |] then
          begin
            if !debug then
              Format.eprintf "or [%a] ; [%a]: closed left side@." (Value.output env) a (Value.output env) b;
            anchors_b, close_b
          end
        else if anchors_b = [| |] then
          begin
            if !debug then
              Format.eprintf "or [%a] ; [%a]: closed right side@." (Value.output env) a (Value.output env) b;
            anchors_a, close_a
          end
        else if !m then
          begin
            if !debug then
              Format.eprintf "or [%a] ; [%a]: divergence@." (Value.output env) a (Value.output env) b;
            let close nodes =
              let length_a = Array.length anchors_a in
              let length_b = Array.length anchors_b in
              mk_or pos [close_a (Array.sub nodes 0 length_a); close_b (Array.sub nodes length_a length_b)] in
            Array.append anchors_a anchors_b, close
          end
        else
          begin
            if !debug then
              Format.eprintf "or [%a] ; [%a]: joined@." (Value.output env) a (Value.output env) b;
            let close nodes =
              let tree_a = close_true pos (anchors_a, close_a) in
              let tree_b = close_true pos (anchors_b, close_b) in
              mk_and pos [mk_or pos [tree_a; tree_b]; nodes.(0)] in
            [| env |], close
          end
    | Value.Term ("=", [| a; b |]), pos ->
        eval_unify pos a b env []
    | Value.Term ("\\=", [| a; b |]), pos ->
        eval history env k_exp (Value.Term ("\\+", [| Value.Term ("=", [| a; b |]), pos |]), pos)
    | Value.Term ("==", [| a; b |]), pos ->
        check_equal pos a b
    | Value.Term ("\\==", [| a; b |]), pos ->
        check_not_equal pos a b
    | Value.Term ("<=", [| a; b |]), pos ->
        check_order pos "<=" ( <= ) ( <= ) a b
    | Value.Term (">", [| a; b |]), pos ->
        check_order pos ">" ( > ) ( > ) a b
    | Value.Term (">=", [| a; b |]), pos ->
        check_order pos ">=" ( >= ) ( >= ) a b
    | Value.Term ("<", [| a; b |]), pos ->
        check_order pos "<" ( < ) ( < ) a b
    | Value.Term (f, args), pos when
        f = "int_search" && Array.length args = 4 ||
        (f = "store" || f = "retrieve") && Array.length args = 2 ||
        (f = "minimize" || f = "maximize") && Array.length args = 1 ||
        f = "satisfy" && Array.length args = 0 ->
        let orders, args = eval_args None pos args in
        let args = Array.map (Ground.of_value env) args in  
        let s = Ground.Term (f, args), pos in
        goal_seq pos env (orders @ [Search s, pos])
    | Value.Term ("domain", [| x; a; b |]), pos ->
        let d0 = fresh_var () in
        let orders_a, a = eval_arithmetic domain_table None env a in
        let orders_b, b = eval_arithmetic domain_table None env b in
        begin
          try
            let a = Value.to_int_conv env a and b = Value.to_int_conv env b in
            StringHashtbl.add domain_table d0 (Domain.Int (a, b))
          with Value.Conversion ->
            let a = Value.to_float env a and b = Value.to_float env b in
            StringHashtbl.add domain_table d0 (Domain.RangeFloat (a, b))
        end;
        eval_unify pos x (Value.Term (d0, [| |]), pos) env (make_search pos (orders_a @ orders_b))
    | Value.Term ("print", [| v |]), pos ->
        let orders, v = eval_arithmetic domain_table None env v in
        print_string (Value.to_string env v);
        goal_seq pos env (make_search pos orders)
    | Value.Term ("nl", [| |]), _ ->
        print_newline ();
        goal_true env
    | Value.Term ("time", [| t |]), pos ->
        let orders, t = eval_arithmetic domain_table None env t in
        eval_unify pos t (Value.Float (Sys.time ()), pos) env (make_search pos orders)
    | Value.Term ("g_assign", [| x; v |]), _ ->
        let x = Value.to_atom env x in
        let v = Value.deep_deref env v in
        StringHashtbl.replace global_vars x v;
        goal_true env
    | Value.Term ("g_assignb", [| x; v |]), _ ->
        let x' =
          match Value.fresh () with
          | Value.Var (x', _), _ -> x'
          | _ -> failwith "var" in
        let x = Value.to_atom env x in
        StringHashtbl.replace global_bvars x x';
        let env = IntMap.add x' v env in
        goal_true env
    | Value.Term ("g_read", [| x; v |]), pos ->
        let x = Value.to_atom env x in
        let v' =
          begin
            try
              IntMap.find (StringHashtbl.find global_bvars x) env
            with Not_found ->
              try
              StringHashtbl.find global_vars x
              with Not_found ->
                (Value.Int 0, pos)
          end in
        eval_unify pos v v' env []
    | Value.Term (f, args), pos ->
        let arity = Array.length args in
        let f_clauses =
          try
            Some (StringIntHashtbl.find clauses (f, arity))
          with Not_found ->
            None in
        let orders, args = eval_args (Some clauses) pos args in
        let t = (Value.Term (f, args), pos) in
        begin
          match f_clauses with
          | None ->
              let orders', t = eval_arithmetic domain_table (Some clauses) env t in
              let orders = orders @ make_search pos orders' in
              goal_seq pos env (orders @ [Constraint (Ground.of_value env t, false), pos])
          | Some f_clauses ->
              if (List.exists (same_goal domain_table env t) history) then
                Lexer.error pos ("infinite recursion on " ^ Value.to_string env t);
              let anchors, close =
                eval (t :: history) env k_exp (goal_of_clauses f args f_clauses pos) in
              let close nodes =
                mk_and pos (orders @ [close nodes]) in
              anchors, close
        end
    | _, pos ->
        Lexer.error pos "goal expected" in
  close_true dummy_pos (eval [] IntMap.empty (Value.Term ("true", [| |]), dummy_pos) goal)
    
let rec negate_constraint cstr =
  match cstr with
  | Ground.Term ("false", [| |]), pos ->
      Ground.Term ("true", [| |]), pos
  | Ground.Term ("true", [| |]), pos ->
      Ground.Term ("false", [| |]), pos
  | Ground.Term ("=", [| a; b |]), pos ->
      Ground.Term ("!=", [| a; b |]), pos
  | Ground.Term ("!=", [| a; b |]), pos ->
      Ground.Term ("=", [| a; b |]), pos
  | Ground.Term ("<", [| a; b |]), pos ->
      Ground.Term (">=", [| a; b |]), pos
  | Ground.Term ("<=", [| a; b |]), pos ->
      Ground.Term (">", [| a; b |]), pos
  | Ground.Term (">", [| a; b |]), pos ->
      Ground.Term ("<=", [| a; b |]), pos
  | Ground.Term (">=", [| a; b |]), pos ->
      Ground.Term ("<", [| a; b |]), pos
  | Ground.Term ("/\\", [| a; b |]), pos ->
      Ground.Term ("\\/", [| negate_constraint a; negate_constraint b |]), pos
  | Ground.Term ("\\/", [| a; b |]), pos ->
      Ground.Term ("/\\", [| negate_constraint a; negate_constraint b |]), pos
  | Ground.Term ("not", [| a |]), pos ->
      a
  | _, pos ->
      Ground.Term ("not", [| cstr |]), pos

let list_of_and tree =
  match tree with
  | And list, _ ->
      list
  | _ ->
      [tree]

let rec remove_constraint cstr list =
  match list with
  | [] -> raise Not_found
  | (Constraint (hd, _), _) :: tl when Ground.compare cstr hd = 0 ->
      tl
  | hd :: tl -> hd :: remove_constraint cstr tl
        
let rec recognize_equiv tree =
  match tree with
  | And list, pos ->
      And (List.map recognize_equiv list), pos
  | Or list, pos ->
      let list = List.map recognize_equiv list in
      let branchs = List.map list_of_and list in
      let rec check_branchs accu branchs =
        match branchs with
        | [] -> List.rev accu
        | branch :: branchs ->
            let check_cstr (accu, branchs, accu') tree =
              match tree with
              | Constraint (cstr, flag), pos ->
                  let neg = negate_constraint cstr in
                  begin
                    try
                      let accu = List.map (remove_constraint neg) accu in
                      let branchs = List.map (remove_constraint neg) branchs in
                      (accu, branchs, (Constraint (cstr, true), pos) :: accu')
                    with Not_found ->
                      (accu, branchs, tree :: accu')
                  end
              | _ ->
                  (accu, branchs, tree :: accu') in
            let (accu, branchs, accu') =
              List.fold_left check_cstr (accu, branchs, []) branch in
            check_branchs (List.rev accu' :: accu) branchs in
      let branchs = check_branchs [] branchs in
      mk_or pos (List.map (mk_and pos) branchs)
  | _ ->
      tree

let objective = ref None
        
let generate_csp domain_table tree =
  let or_vars = ref [| |] in
  let introduce_or_var or_depth size =
    let orig_length = Array.length !or_vars and min_length = or_depth + 1 in
    let is_new = orig_length < min_length in
    if is_new then
      or_vars := Array.append !or_vars [| (fresh_var (), 0) |];
    let or_vars = !or_vars in
    let (var_name, orig_size) = or_vars.(or_depth) in
    if orig_size < size then
      or_vars.(or_depth) <- (var_name, size);
    is_new, var_name in
  let build_node pos (d, i) =
    Ground.Term ("=", [| Ground.Term (d, [| |]), pos; Ground.Int i, pos |]), pos in
  let build_branch pos branch =
    let add_branch cstr node =
      Ground.Term ("/\\", [| build_node pos node; cstr |]), pos in
    List.fold_left add_branch (Ground.Term ("true", [| |]), pos) branch in
  let reject_table = StringHashtbl.create 17 in
  let add_reject d branch_cstr =
    try
      StringHashtbl.replace reject_table d (branch_cstr :: StringHashtbl.find reject_table d)
    with Not_found ->
      StringHashtbl.add reject_table d [branch_cstr] in
  let rec generate branch (constraints, annotations, or_depth) tree =
    match tree with
    | False, pos ->
        Lexer.error pos "false in decision tree"
    | True, _ ->
        (constraints, annotations, or_depth)
    | And list, _ ->
        List.fold_left (generate branch) (constraints, annotations, or_depth) list
    | Or list, pos ->
        let size = List.length list in
        let is_new, d = introduce_or_var or_depth size in
        let or_depth = succ or_depth in
        let branch_cstr = build_branch pos branch in
        let d_bound = Ground.Term ("->", [| branch_cstr; (Ground.Term ("<", [| Ground.Term (d, [| |]), pos; Ground.Int size, pos |]), pos) |]), pos in
        add_reject d branch_cstr;
        let constraints = d_bound :: constraints in
        let annotations =
          if is_new then
            let annotation = Ground.Term ("int_search", [| Ground.Term (".", [| Ground.Term (d, [| |]), pos; Ground.Term ("[]", [| |]), pos |]), pos; Ground.Term ("input_order", [| |]), pos; Ground.Term ("indomain_min", [| |]), pos; Ground.Term ("complete", [| |]), pos|]), pos in
            annotation :: annotations
          else
            annotations in
        let generate' (constraints, annotations, max_or_depth, index) child =
          let branch = (d, index) :: branch in
          let (constraints, annotations, or_depth') = generate branch (constraints, annotations, or_depth) child in
          (constraints, annotations, max max_or_depth or_depth', succ index) in
        let (constraints, annotations, or_depth, _) =
          List.fold_left generate' (constraints, annotations, or_depth, 0) list in
        (constraints, annotations, or_depth)
    | Search s, pos ->
        let constraints, s =
          match s with
          | Ground.Term ("store", [| global; source |]), pos ->
              let v = fresh_var () in
              StringHashtbl.add domain_table v Domain.Bool;
              let b = Ground.Term (v, [| |]), pos in
              (Ground.Term ("<->", [| b; build_branch pos branch |]), pos) :: constraints,
              (Ground.Term ("store", [| b; global; source |]), pos)
          | _ ->
              if branch <> [] then
                begin
                  match s with
                  | Ground.Term (f, [| _; _ |]), _ when f = "indexical_min" || f = "indexical_max" || f = "indexical_card" || f = "retrieve" ->
                      ()
                  | Ground.Term ("indexical_dom_nth", [| _; _; _ |]), _ ->
                      ()
                  | _ ->
                      Lexer.error pos "search annotation unallowed below choice-points"
                end;
              constraints, s in
        begin
          match s with
          | Ground.Term (("minimize" | "maximize"), [| _ |]), _
          | Ground.Term (("satisfy"), [| |]), _ ->
              begin
                match !objective with
                | None ->
                    objective := Some s
                | Some _ ->
                    Lexer.error pos "only one objective allowed"
              end;
              (constraints, annotations, or_depth)
          | _ ->
              (constraints, s :: annotations, or_depth)
        end
    | Constraint (c, equiv), pos ->
        let cstr =
          if equiv then
            match branch with
            | node :: branch ->
                let branch_cstr = build_branch pos branch in
                (Ground.Term ("->", [| branch_cstr; Ground.Term ("<->", [| build_node pos node; c |]), pos |]), pos)
            | _ ->
                failwith "equiv"
          else
            let branch_cstr = build_branch pos branch in
            (Ground.Term ("->", [| branch_cstr; c |]), pos) in
        (cstr :: constraints, annotations, or_depth) in
  let (constraints, annotations, _) =
    generate [] ([], [], 0) tree in
  let add_or_domain (x, size) =
    StringHashtbl.add domain_table x (Domain.Int (0, size - 1)) in
  Array.iter add_or_domain !or_vars;
  let add_reject_constraint d branches constraints =
    let add_branch cstr branch_cstr =
      Ground.Term ("/\\", [| cstr; Ground.Term ("not", [| branch_cstr |]), dummy_pos |]), dummy_pos in
    let conj = List.fold_left add_branch (Ground.Term ("true", [| |]), dummy_pos) branches in
    let d_reject = Ground.Term ("->", [| conj; (Ground.Term ("=", [| Ground.Term (d, [| |]), dummy_pos; Ground.Int 0, dummy_pos |]), dummy_pos) |]), dummy_pos in
    d_reject :: constraints in
  let constraints = StringHashtbl.fold add_reject_constraint reject_table constraints in
  (List.rev constraints, List.rev annotations)

let simplify_expression domain_table equations expr =
  let rec simplify_expression expr =
    let expr =
      match expr with
      | Ground.Term (f, args), pos ->
          Ground.Term (f, Array.map simplify_expression args), pos
      | _ ->
          expr in
    match expr with
    | Ground.Term (x, [| |]), pos ->
        begin
          try
            simplify_expression (StringHashtbl.find equations x)
          with Not_found ->
            expr
        end
    | Ground.Term ("=", [| Ground.Int i, _; Ground.Int j, _ |]), pos ->
        if i = j then
          Ground.Term ("true", [| |]), pos
        else
          Ground.Term ("false", [| |]), pos
    | Ground.Term ("=", [| a; b |]), pos ->
        if Ground.compare a b = 0 then
          Ground.Term ("true", [| |]), pos
        else
          expr
    | Ground.Term ("!=", [| Ground.Term (x, [| |]), _; Ground.Int i, _ |]), pos
    | Ground.Term ("!=", [| Ground.Int i, _; Ground.Term (x, [| |]), _ |]), pos ->
        begin
          try
            match StringHashtbl.find domain_table x with
            | Domain.Int (lb, ub) ->
                if i < lb || ub < i then
                  Ground.Term ("true", [| |]), pos
                else if ub = lb + 1 then
                  if i = lb then
                    Ground.Term ("=", [| Ground.Term (x, [| |]), pos; Ground.Int ub, pos |]), pos
                  else
                    Ground.Term ("=", [| Ground.Term (x, [| |]), pos; Ground.Int lb, pos |]), pos
                else
                  expr
            | _ ->
                expr
          with Not_found ->
            expr
        end
    | Ground.Term ("->", [| a; b |]), pos ->
        begin
          match a, b with
          | _, (Ground.Term ("true", [| |]), _)
          | (Ground.Term ("true", [| |]), _), _ ->
              b
          | (Ground.Term ("false", [| |]), _), _ ->
              Ground.Term ("true", [| |]), pos
          | _, (Ground.Term ("false", [| |]), _) ->
              simplify_expression (Ground.Term ("not", [| a |]), pos)
          | _ ->
              Ground.Term ("->", [| a; b |]), pos
        end
    | Ground.Term ("<->", [| a; b |]), pos ->
        begin
          match a, b with
          | _, (Ground.Term ("true", [| |]), _) ->
              a
          | (Ground.Term ("true", [| |]), _), _ ->
              b
          | _ ->
              Ground.Term ("<->", [| a; b |]), pos
        end
    | Ground.Term ("/\\", [| a; b |]), pos ->
        begin
          match a, b with
          | (Ground.Term ("true", [| |]), _), _ ->
              b
          | _, (Ground.Term ("true", [| |]), _) ->
              a
          | _ ->
              Ground.Term ("/\\", [| a; b |]), pos
        end
    | Ground.Term ("\\/", [| a; b |]), pos ->
        begin
          match a, b with
          | (Ground.Term ("true", [| |]), _), _
          | _, (Ground.Term ("true", [| |]), _) ->
              Ground.Term ("true", [| |]), pos
          | (Ground.Term ("false", [| |]), _), _ ->
              b
          | _, (Ground.Term ("false", [| |]), _) ->
              a
          | _ ->
              Ground.Term ("\\/", [| a; b |]), pos
        end
    | Ground.Term ("not", [| a |]), pos ->
        let a = negate_constraint a in
        let a =
          match a with
          | Ground.Term ("not", [| _ |]), _ -> a
          | a -> simplify_expression a in
        a
    | Ground.Term ("<", [| Ground.Term (x, [| |]), _; Ground.Int ub, _ |]), pos ->
        begin
          try
            let min, max = Domain.get_int pos (StringHashtbl.find domain_table x) in
            if ub > max then
              Ground.Term ("true", [| |]), pos
            else
              expr
          with Not_found ->
            expr
        end
    | Ground.Term ("+", [| a; b |]), pos ->
        begin
          match a, b with
          | (Ground.Int 0, _), _ ->
              b
          | _, (Ground.Int 0, _) ->
              a
          | _ ->
              Ground.Term ("+", [| a; b |]), pos
        end
    | Ground.Term ("*", [| a; b |]), pos ->
        begin
          match a, b with
          | (Ground.Int 0, _), _ ->
              (Ground.Int 0, pos)
          | (Ground.Int 1, _), _ ->
              b
          | _, (Ground.Int 0, _) ->
              (Ground.Int 0, pos)
          | _, (Ground.Int 1, _) ->
              a
          | _ ->
              Ground.Term ("*", [| a; b |]), pos
        end
    | Ground.Term ("div", [| a; b |]), pos ->
        begin
          match a, b with
          | _, (Ground.Int 1, _) ->
              a
          | _ ->
              Ground.Term ("div", [| a; b |]), pos
        end
    | _ -> expr in
  simplify_expression expr

let suppress domain_table equations x b =
  if StringHashtbl.mem domain_table x then
    begin
      StringHashtbl.add equations x b;
      StringHashtbl.remove domain_table x
    end
        
let infer_equations domain_table equations constraints =
  let rec occur x b =
    match b with
    | Ground.Term (y, [| |]), _ ->
        x = y
    | Ground.Term (_, args), _ ->
        List.exists (occur x) (Array.to_list args)
    | _ ->
        false in
  let rec make_zero sum =
    match sum with
    | Ground.Term ("+", [| a; b |]), _ ->
        make_zero a;
        make_zero b
    | Ground.Term (x, [| |]), pos ->
        suppress domain_table equations x (Ground.Int 0, pos)
    | _ ->
        () in
  let rec positive_sum sum =
    match sum with
    | Ground.Term ("+", [| a; b |]), _ ->
        positive_sum a && positive_sum b
    | Ground.Term (x, [| |]), _ when x.[0] = 'X' ->
        begin
          try
            match StringHashtbl.find domain_table x with
            | Domain.Int (0, _) ->
                true
            | _ ->
                false
          with Not_found ->
            false
        end
    | Ground.Int 0, _ ->
        true
    | _ ->
        false in
  let infer_equations cstr =
    match cstr with
    | Ground.Term ("=", [| Ground.Term (x, [| |]), _; b |]), pos
      when not (occur x b) && x.[0] = 'X' ->
        suppress domain_table equations x b
    | Ground.Term ("=", [| b; Ground.Term (x, [| |]), _ |]), pos
      when not (occur x b) && x.[0] = 'X'  ->
        suppress domain_table equations x b
    | Ground.Term ("=", [| Ground.Int 0, _; sum |]), pos
      when positive_sum sum  ->
        make_zero sum
    | Ground.Term ("=", [| sum; Ground.Int 0, _ |]), pos
      when positive_sum sum  ->
        make_zero sum
    | _ ->
        () in
  List.iter infer_equations constraints

let not_instanciated v =
  match v with
  | Ground.Int _, _ ->
      false
  | _ ->
      true
        
let simplify_annotations domain_table equations annotations =
  let simplify_annotation annotation =
    match annotation with
    | Ground.Term ("int_search", [| Ground.Term (".", [| _; _ |]), _ as vars; var_sel; val_sel; mode |]), pos ->
        let vars = Ground.to_list vars in
        let vars = List.filter not_instanciated vars in
        if vars = [] then
          []
        else
          [Ground.Term ("int_search", [| Ground.of_list pos vars; var_sel; val_sel; mode |]), pos]
    | Ground.Term ("indexical_min", [| Ground.Term (target, [| |]), _; (Ground.Int i, _) as x |]), _ ->
        suppress domain_table equations target x;
        []
    | Ground.Term ("indexical_max", [| Ground.Term (target, [| |]), _; (Ground.Int i, _) as x |]), _ ->
        suppress domain_table equations target x;
        []
    | Ground.Term ("indexical_card", [| Ground.Term (target, [| |]), _; Ground.Int _, _ |]), pos ->
        suppress domain_table equations target (Ground.Int 1, pos);
        []
    | Ground.Term ("indexical_dom_nth", [| Ground.Term (target, [| |]), _; Ground.Int i, _; _ |]), pos ->
        suppress domain_table equations target (Ground.Int i, pos);
        []
    | _ ->
        [annotation] in
  List.flatten (List.map simplify_annotation annotations)
    
let simplify_constraints domain_table constraints annotations =
  let is_not_true cstr =
    match cstr with
    | Ground.Term ("true", [| |]), _ ->
        false
    | _ ->
        true in
  let equations = StringHashtbl.create 17 in
  let constraints = List.filter is_not_true (List.map (simplify_expression domain_table equations) constraints) in
  let rec simplify_loop constraints annotations =
    let equations = StringHashtbl.create 17 in
    infer_equations domain_table equations constraints;
    let annotations = simplify_annotations domain_table equations annotations in
    let constraints = List.filter is_not_true (List.map (simplify_expression domain_table equations) constraints) in
    let annotations = List.map (simplify_expression domain_table equations) annotations in
    if StringHashtbl.length equations > 0 then
      simplify_loop constraints annotations
    else
      (constraints, annotations) in
  simplify_loop constraints annotations
    
let add_definition domain_table clauses entry =
  match entry with
  | Clause (f, args, ((_, pos) as body)) ->
      asserta clauses f args body;
      asserta clauses "clause" [| (Term(f, args), pos); body |] ((Term ("true", [| |]), pos))
  | Var_decl decl ->
      begin
        let ty, ((_, pos) as var) = expect_colon decl in
        let var = expect_ident var in
        if var.[0] = 'X' then
          Lexer.error pos "identifiers beginning with an X are reserved for clp2zinc";
        let eval_number e =
          snd (eval_arithmetic domain_table None IntMap.empty (value_of_term (StringHashtbl.create 1) e)) in
        let eval_int i =
          Value.to_int IntMap.empty (eval_number i) in
        let eval_int_conv i =
          Value.to_int_conv IntMap.empty i in
        let eval_float f =
          Value.to_float IntMap.empty f in
        let bound_domain lb ub =
          let lb = eval_number lb in
          let ub = eval_number ub in
          begin
            try
              Domain.Int (eval_int_conv lb, eval_int_conv ub)
            with Value.Conversion ->
              Domain.RangeFloat (eval_float lb, eval_float ub)
          end in
        let extract_domain domain =
          match domain with
          | Term ("..", [| lb; ub |]), _ ->
              bound_domain lb ub
          | Term ("bool", [| |]), _ ->
              Domain.Bool
          | Term ("float", [| |]), _ ->
              Domain.Float
          | _ ->
              raise Exit in
        let rec extract_ty ty =
          match ty with
          | Term ("of", [| Term ("[]", [| Term("array", [| |]), _; Term ("..", [| lb; ub |]), _ |]), _; ty |]), _ ->
              Domain.Array (eval_int lb, eval_int ub, extract_ty ty)
          | Term ("of", [| Term ("[]", [| Term("array", [| |]), _; Term ("..", [| lb0; ub0 |]), _; Term ("..", [| lb1; ub1 |]), _ |]), _; ty |]), _ ->
              Domain.Matrix (eval_int lb0, eval_int ub0, eval_int lb1, eval_int ub1, extract_ty ty)
          | Term ("var", [| dom |]), _ ->
              extract_domain dom
          | _ ->
              raise Exit in
        try StringHashtbl.add domain_table var (extract_ty ty)
        with Exit -> ()
      end
  | _ ->
      ()

let make_atomic domain_table arg =
  match arg with
  | Ground.Term (_, [| |]), _ | Ground.Int _, _ | Ground.Float _, _ ->
      [], arg
  | _, pos ->
      let x = fresh_var () in
      let (lb, ub) = (-10000, 10000) in
      StringHashtbl.add domain_table x (Domain.Int (lb, ub));
      let v = Ground.Term (x, [| |]), pos in
      [Ground.Term ("=", [| v; arg |]), pos], v

let atomic_annotation domain_table annotation =
  match annotation with
  | Ground.Term ("store", [| cond; target; values |]), pos ->
      let values = Ground.to_list values in
      let constraints_values = List.map (make_atomic domain_table) values in
      let constraints' = List.flatten (List.map fst constraints_values) in
      let values = Ground.of_list pos (List.map snd constraints_values) in
      constraints', (Ground.Term ("store", [| cond; target; values |]), pos)
  | Ground.Term ("indexical_dom_nth", [| target; x; n |]), pos ->
      let constraints, n = make_atomic domain_table n in
      constraints, (Ground.Term ("indexical_dom_nth", [| target; x; n |]), pos)
  | _ ->
      [], annotation
        
let atomic_annotations domain_table annotations =
  let constraints_annotations =
    List.map (atomic_annotation domain_table) annotations in
  (List.flatten (List.map fst (constraints_annotations)),
   List.map snd constraints_annotations)

let count = ref false

let interpret domain_table clauses tree_output out_chan entry =
  match entry with
  | Goal ((_, pos) as goal) ->
      let goal = value_of_term (StringHashtbl.create ctx_sz) goal in
      if !debug then
        Format.eprintf "evaluate %a@." (Value.output IntMap.empty) goal;
      let tree = evaluate_goal domain_table clauses goal in
      if tree_output then
        Format.fprintf out_chan "\\Tree %a\n" output_decision_tree tree
      else
        let tree = recognize_equiv tree in
        let (constraints, annotations) = generate_csp domain_table tree in
        let (constraints, annotations) = simplify_constraints domain_table constraints annotations in
        let (constraints', annotations) = atomic_annotations domain_table annotations in
        let constraints = constraints @ constraints' in
        if !count then
          Printf.fprintf stderr "%d constraints\n" (List.length constraints);
        let count = ref 0 in
        let print_dom x dom =
          if x.[0] = 'X' then
            begin
              if !count > 0 then
                if !count mod 3 = 0 then
                  Format.fprintf out_chan "\n"
                else
                  Format.fprintf out_chan "  ";
              incr count;
              Format.fprintf out_chan "%a: %s;" Domain.output dom x
            end in
        StringHashtbl.iter print_dom domain_table;
        if !count >= 1 then
          Format.fprintf out_chan "\n";
        List.iter (fun cstr -> Format.fprintf out_chan "constraint %a;\n" Ground.output cstr) constraints;
        let objective =
          match !objective with
          | None -> Ground.Term ("satisfy", [| |]), dummy_pos
          | Some objective -> objective in
        begin
          match annotations with
          | [] ->
              Format.fprintf out_chan "solve %a;\n" Ground.output objective
          | annotation :: annotations ->
              Format.fprintf out_chan "solve :: seq_search([\n    %a" Ground.output annotation;
              List.iter (fun annotation -> Format.fprintf out_chan ",\n    %a" Ground.output annotation) annotations;
              Format.fprintf out_chan "\n  ]) %a;\n" Ground.output objective
        end        
  | Minizinc minizinc ->
      if tree_output then
        ()
      else
        begin
          Format.fprintf out_chan "%s\n" minizinc;
        end
  | Var_decl decl ->
      if tree_output then
        ()
      else
        Format.fprintf out_chan "%a;\n" (Value.output IntMap.empty) (value_of_term (StringHashtbl.create 1) decl)
  | _ ->
      ()
        
let main () =
  let file_queue = Queue.create () in
  let output_filename_ref = ref None in
  let add_file filename =
    Queue.add filename file_queue in
  let set_output filename =
    output_filename_ref := Some filename in
  let add_prefix prefix =
    prefix_path := prefix :: !prefix_path in
  let add_search_path path =
    search_path := path :: !search_path in
  let tree_output = ref false in
  let options = [
    "-e", Arg.Set expand_rules, "expand rules";
    "-g", Arg.Set debug, "debug";
    "-t", Arg.Set tree_output, "output tree";
    "-o", Arg.String set_output, "set output filename";
    "-O", Arg.String add_prefix, "add prefix to search paths";
    "-I", Arg.String add_search_path, "add search path";
    "--count", Arg.Set count, "count constraints"] in
  Arg.parse options add_file "clp2zinc [FILES]";
  try
    let files = list_of_queue file_queue in
    let first_file =
      match files with
      | [] -> failwith "No file to compile"
      | head :: _ -> head in
    let output_filename =
      match !output_filename_ref with
      | Some filename -> filename
      | None ->
          let output_ext =
            if !tree_output then ".inc"
            else ".mzn" in
          let basename =
            try Filename.chop_extension first_file
            with Invalid_argument _ -> first_file in
          basename ^ output_ext in
    let entries = List.flatten (List.map load_file files) in
    let clauses = StringIntHashtbl.create clause_table_sz in
    let domain_table = StringHashtbl.create domain_table_sz in
    List.iter (add_definition domain_table clauses) (List.rev entries);
    let out_chan = open_out output_filename in
    try
      let fmt = Format.formatter_of_out_channel out_chan in
      List.iter (interpret domain_table clauses !tree_output fmt) entries;
      Format.pp_print_flush fmt ();
      close_out out_chan
    with err ->
      close_out_noerr out_chan;
      raise err
  with
  | Failure msg
  | Sys_error msg ->
      prerr_endline msg;
      exit 1
  | Lexer.Error (position, msg) ->
      if position.Lexer.start_p.Lexer.line = position.Lexer.end_p.Lexer.line then
        if position.Lexer.start_p.Lexer.column = position.Lexer.end_p.Lexer.column then
          Format.eprintf "%s:%d.%d: %s@." position.Lexer.filename position.Lexer.start_p.Lexer.line position.Lexer.start_p.Lexer.column msg
        else
          Format.eprintf "%s:%d.%d-%d: %s@." position.Lexer.filename position.Lexer.start_p.Lexer.line position.Lexer.start_p.Lexer.column position.Lexer.end_p.Lexer.column msg
      else
        Format.eprintf "%s:%d.%d-%d.%d: %s@." position.Lexer.filename position.Lexer.start_p.Lexer.line position.Lexer.start_p.Lexer.column position.Lexer.end_p.Lexer.line position.Lexer.end_p.Lexer.column msg;
      exit 1

let () =
  if not !Sys.interactive then
    main ()
